function SetTemplateParameters
% SetTemplateParameters - Set parameters for the image template.
% 
%   Parameters are hard coded in this .m file. 
%   Please directly edit the code.
%
%
%   Gabor feature parameters:
%
%         Scales and the corresponding filter sizes
%         0.2: 5 * 5
%         0.3: 9 * 9
%         0.4: 11 * 11
%         0.6: 15 * 15
%         0.8: 21 * 21
%         1.2: 29 * 29
%         1.6: 39 * 39

addpath('common');

%% Feature parameters

%
% Use single precision numbers to save memory.
%
SinglePrecision = 1;

%
% Gabor scale, #orientations
%
GaborScale = single([0.2, 0.3, 0.4, 0.6, 0.8, 1.2, 1.6]);
nGaborScale = length(GaborScale);
nGaborOri = 16;
GaborFilter = cell(nGaborScale*nGaborOri,1);
GaborSymbol = cell(nGaborScale*nGaborOri,1);
for iScale = 1:nGaborScale
    [ GaborFilter( (iScale-1)*nGaborOri + (1:nGaborOri) ) ...
        GaborSymbol( (iScale-1)*nGaborOri + (1:nGaborOri) ) ]...
        = makefilter(GaborScale(iScale), nGaborOri, SinglePrecision);
end

%
% Correlation betweeen Gabor filters. Used for local inhibition.
%
CorrGabor = cell(nGaborOri*nGaborScale, nGaborOri*nGaborScale);
GaborOverlapTolerance = 0.1;
for iF = 1:nGaborOri*nGaborScale
    hi = (size(GaborFilter{iF}, 1) - 1) / 2;            % half size of filter iF
    for jF = 1:nGaborOri*nGaborScale
        hj = (size(GaborFilter{jF}, 1) - 1) / 2;        % half size of filter jF
        I = zeros(2*(hi+hj)+1)+sqrt(-1)*zeros(2*(hi+hj)+1);
        I((hj+1):(hj+2*hi+1), (hj+1):(hj+2*hi+1)) = GaborFilter{iF}; % put GaborFilter{iF} at the center   
        % convolve GaborFilter{iF} by GaborFilter{jF}, 4 real-imaginary combinations
        rr = filter2(real(GaborFilter{jF}), real(I), 'same');
        ri = filter2(real(GaborFilter{jF}), imag(I), 'same');
        ir = filter2(imag(GaborFilter{jF}), real(I), 'same');
        ii = filter2(imag(GaborFilter{jF}), imag(I), 'same');
        if SinglePrecision > 0
            CorrGabor{iF, jF} = single( ((rr.*rr + ri.*ri + ir.*ir + ii.*ii)<GaborOverlapTolerance)*1.0 );
        else
            CorrGabor{iF, jF} = double( ((rr.*rr + ri.*ri + ir.*ir + ii.*ii)<GaborOverlapTolerance)*1.0 );
        end
        % CorrGabor{i,j} is how filter i at (0,0) inhibites filter j at locations overlapping with filter i
   end
end


%
% Parameter for Gabor filter responses
%
GaborSaturation = 6;

%
% Parameter for exponential model
%
binSize = .2;  % binsize for computing histogram of q()
numStoredPoint = 40; % number of stored lambda values
spacing = .05; % spacing between two adjacent points

% save
save('template_params.mat');


